/**************************************************************************

	ns

	script_parse.c

	(C) Copyright 1996-2001 By Tomoaki Nakashima. All right reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>

#include "script.h"
#include "script_regex.h"
#include "script_string.h"
#include "script_memory.h"


/**************************************************************************
	Define
**************************************************************************/

#define IS_SPACE(c)				(c == ' ' || c == '\t' || c == '\r' || c == '\n')


/**************************************************************************
	Global Variables
**************************************************************************/

typedef struct _PARSEINFO {
	EXECINFO *ei;
	int type;
	int level;
	char *p;
	char *r;
	BOOL concat;
	BOOL decl;
	BOOL postfix;
} PARSEINFO;


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

static BOOL GetToken(PARSEINFO *pi);
static TOKEN *CreateToken(int type, char *p);

//
static TOKEN *Primary(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Hash(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *UnaryOperator(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Multiplicative(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Additive(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Shift(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Relational(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Equality(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *AND(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *ExclusiveOR(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *OR(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *LogicalAND(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *LogicalOR(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Assignment(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Postfix(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *Expression(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *ExpressionStatement(PARSEINFO *pi, TOKEN *cu_tk);

//
static TOKEN *Condition(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *IfStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *SwitchStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *WhileStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *DoWhileStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *ForStatement(PARSEINFO *pi, TOKEN *cu_tk);

static TOKEN *JumpStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *LabeldStatement(PARSEINFO *pi, TOKEN *cu_tk);

//`
static TOKEN *VarDeclList(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *FuncDecl(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *VarDecl(PARSEINFO *pi, TOKEN *cu_tk);

//\
static TOKEN *CompoundStatement(PARSEINFO *pi, TOKEN *cu_tk);
static TOKEN *StatementList(PARSEINFO *pi, TOKEN *cu_tk);


/******************************************************************************

	GetToken

	

******************************************************************************/

static BOOL GetToken(PARSEINFO *pi)
{
	char *r, *s;

	for(pi->p = pi->r; IS_SPACE(*pi->p); pi->p++);
	pi->r = pi->p;
	pi->type = 0;

	switch(*pi->p)
	{
	case '\0':
		pi->type = SYM_EOF;
		return TRUE;

	case '#':
		pi->type = SYM_PREP;
		pi->concat = TRUE;
		break;

	case ';':
		pi->type = SYM_LINEEND;
		pi->concat = TRUE;
		break;

	case ',':
		pi->type = SYM_WORDEND;
		pi->concat = TRUE;
		break;

	case ':':
		pi->type = SYM_CASEEND;
		pi->concat = TRUE;
		break;

	case '\"':
		if(pi->concat == FALSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_STRING;
		pi->r = str_skip(pi->p);
		if(*pi->r == '\0'){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->concat = FALSE;
		break;

	case '{':
		if(pi->concat == FALSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(get_pair_brace(pi->p) == NULL){
			Error(pi->ei, ERR_BRACE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_BOPEN;
		pi->concat = TRUE;
		break;

	case '}':
		pi->type = SYM_BCLOSE;
		pi->concat = FALSE;
		break;

	case '(':
		if(pi->concat == FALSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(get_pair_brace(pi->p) == NULL){
			Error(pi->ei, ERR_BRACE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_OPEN;
		pi->concat = TRUE;
		break;

	case ')':
		pi->type = SYM_CLOSE;
		pi->concat = FALSE;
		break;

	case '[':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(get_pair_brace(pi->p) == NULL){
			Error(pi->ei, ERR_BRACE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_HASHOPEN;
		pi->concat = TRUE;
		break;

	case ']':
		pi->type = SYM_HASHCLOSE;
		pi->concat = FALSE;
		break;

	case '=':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		switch(*(pi->p + 1))
		{
		case '=':
			pi->type = SYM_EQEQ;
			pi->r++;
			break;

		case '~':
			pi->type = SYM_REGEQ;
			pi->r++;
			break;

		default:
			pi->type = SYM_EQ;
			break;
		}
		pi->concat = TRUE;
		break;

	case '!':
		if(pi->concat == TRUE){
			pi->type = SYM_NOT;
			break;
		}
		if(*(pi->p + 1) == '='){
			pi->type = SYM_NTEQ;
			pi->r++;
		}else{
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->concat = TRUE;
		break;

	case '<':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(*(pi->p + 1) == '<'){
			pi->type = SYM_LEFTSHIFT;
			pi->r++;
		}else if(*(pi->p + 1) == '='){
			pi->type = SYM_LEFTEQ;
			pi->r++;
		}else{
			pi->type = SYM_LEFT;
		}
		pi->concat = TRUE;
		break;

	case '>':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(*(pi->p + 1) == '>'){
			pi->type = SYM_RIGHTSHIFT;
			pi->r++;
		}else if(*(pi->p + 1) == '='){
			pi->type = SYM_RIGHTEQ;
			pi->r++;
		}else{
			pi->type = SYM_RIGHT;
		}
		pi->concat = TRUE;
		break;

	case '&':
		if(pi->concat == TRUE){
			//Xbhs֐
			break;
		}
		if(*(pi->p + 1) == '&'){
			pi->type = SYM_CPAND;
			pi->r++;
		}else{
			pi->type = SYM_AND;
		}
		pi->concat = TRUE;
		break;

	case '|':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		if(*(pi->p + 1) == '|'){
			pi->type = SYM_CPOR;
			pi->r++;
		}else{
			pi->type = SYM_OR;
		}
		pi->concat = TRUE;
		break;

	case '^':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_XOR;
		pi->concat = TRUE;
		break;

	case '~':
		if(pi->concat == FALSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_BITNOT;
		break;

	case '*':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_MULTI;
		pi->concat = TRUE;
		break;

	case '/':
		if(*(pi->p + 1) == '/'){
			//Rg
			for(; *pi->r != '\0' && *pi->r != '\r' && *pi->r != '\n'; pi->r++);
			return GetToken(pi);
		}else if(*(pi->p + 1) == '*'){
			//Rg
			for(; *pi->r != '\0'; pi->r++){
				if(*pi->r == '*' && *(pi->r + 1) == '/'){
					pi->r += 2;
					break;
				}
			}
			return GetToken(pi);
		}
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_DIV;
		pi->concat = TRUE;
		break;

	case '%':
		if(pi->concat == TRUE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return FALSE;
		}
		pi->type = SYM_MOD;
		pi->concat = TRUE;
		break;

	case '+':
		if(pi->concat == TRUE){
			if(*(pi->p + 1) == '+'){
				pi->type = SYM_INC;
				pi->r++;
			}else{
				pi->type = SYM_PLUS;
			}
			break;
		}
		if(*(pi->p + 1) == '+'){
			pi->type = SYM_BINC;
			pi->r++;
			pi->concat = FALSE;
		}else{
			pi->type = SYM_ADD;
			pi->concat = TRUE;
		}
		break;

	case '-':
		if(pi->concat == TRUE){
			if(*(pi->p + 1) == '-'){
				pi->type = SYM_DEC;
				pi->r++;
			}else{
				pi->type = SYM_MINS;
			}
			break;
		}
		if(*(pi->p + 1) == '-'){
			pi->type = SYM_BDEC;
			pi->r++;
			pi->concat = FALSE;
		}else{
			pi->type = SYM_SUB;
			pi->concat = TRUE;
		}
		break;
	}

	if(pi->type != 0){
		pi->r++;
		return TRUE;
	}
	if(pi->concat == FALSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return FALSE;
	}
	pi->concat = FALSE;

	if((*pi->p >= '0' && *pi->p <= '9') ||
		(*pi->p == '.' && *(pi->p + 1) >= '0' && *(pi->p + 1) <= '9')){
		//萔
		pi->type = SYM_CONST;
		if(*pi->p == '0' && (*(pi->p + 1) == 'x' || *(pi->p + 1) == 'X')){
			for(pi->r += 2; (*pi->r >= '0' && *pi->r <= '9') ||
				(*pi->r >= 'A' && *pi->r <= 'F') ||
				(*pi->r >= 'a' && *pi->r <= 'f'); pi->r++);
		}else{
			for(pi->r++; (*pi->r >= '0' && *pi->r <= '9') || *pi->r == '.'; pi->r++);
		}
		return TRUE;
	}

	pi->r = (*pi->p == '&') ? pi->p + 1 : pi->p;
	for(; (*pi->r >= 'a' && *pi->r <= 'z') ||
		(*pi->r >= 'A' && *pi->r <= 'Z') ||
		(*pi->r >= '0' && *pi->r <= '9') ||
		*pi->r == '_' || *pi->r == '.' || *pi->r == '?' ||
		IsDBCSLeadByte((BYTE)*pi->r) == TRUE; pi->r++){
		if(IsDBCSLeadByte((BYTE)*pi->r) == TRUE && *(pi->r + 1) != '\0'){
			//2oCgR[hXLbv
			pi->r++;
			continue;
		}
	}
	if(pi->p == pi->r){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return FALSE;
	}
	s = alloc_copy_n(pi->p, pi->r - pi->p + 1);
	if(s == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return FALSE;
	}

	//\
	if(str_cmp_i(s, "var") == 0){
		pi->type = SYM_VAR;
	}else if(str_cmp_i(s, "return") == 0){
		pi->type = SYM_RETURN;
	}else if(str_cmp_i(s, "resume") == 0){
		pi->type = SYM_RETURN;
	}else if(str_cmp_i(s, "exit") == 0){
		pi->type = SYM_EXIT;
	}else if(str_cmp_i(s, "if") == 0){
		pi->type = SYM_IF;
	}else if(str_cmp_i(s, "else") == 0){
		pi->type = SYM_ELSE;
	}else if(str_cmp_i(s, "for") == 0){
		pi->type = SYM_FOR;
	}else if(str_cmp_i(s, "while") == 0){
		pi->type = SYM_WHILE;
	}else if(str_cmp_i(s, "do") == 0){
		pi->type = SYM_DO;
	}else if(str_cmp_i(s, "break") == 0){
		pi->type = SYM_BREAK;
	}else if(str_cmp_i(s, "continue") == 0){
		pi->type = SYM_CONTINUE;
	}else if(str_cmp_i(s, "switch") == 0){
		pi->type = SYM_SWITCH;
	}else if(str_cmp_i(s, "case") == 0){
		pi->type = SYM_CASE;
	}else if(str_cmp_i(s, "default") == 0){
		pi->type = SYM_DEFAULT;
	}else if(str_cmp_i(s, "function") == 0){
		pi->type = SYM_FUNCTION;
	}else if(str_cmp_i(s, "argmatch") == 0){
		pi->type = SYM_ARGMATCH;
	}
	mem_free(&s);
	if(pi->type != 0){
		pi->concat = TRUE;
		return TRUE;
	}

	//֐ϐ
	for(r = pi->r; IS_SPACE(*r); r++);
	if(*r == '('){
		pi->type = SYM_FUNC;
		pi->concat = TRUE;
	}else{
		pi->type = SYM_VARIABLE;
	}
	return TRUE;
}


/******************************************************************************

	CreateToken

	g[N̍쐬

******************************************************************************/

static TOKEN *CreateToken(int type, char *p)
{
	TOKEN *tk;

	tk = mem_alloc(sizeof(TOKEN));
	if(tk == NULL){
		return NULL;
	}
	ZeroMemory(tk, sizeof(TOKEN));
	tk->Type = type;
	tk->Err = p;
	return tk;
}


/******************************************************************************

	Primary

	vf

******************************************************************************/

static TOKEN *Primary(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk, ttk;
	char *s;

	switch(pi->type)
	{
	case SYM_BOPEN:
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		ttk.next = NULL;
		if(Expression(pi, &ttk) == NULL){
			FreeToken(ttk.next);
			return NULL;
		}
		cu_tk->target = ttk.next;

		if(pi->type != SYM_BCLOSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return NULL;
		}
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		break;

	case SYM_OPEN:
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		cu_tk = Assignment(pi, cu_tk);
		if(pi->type != SYM_CLOSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			return NULL;
		}
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		break;

	case SYM_FUNC:
		//̊Jnʒuݒ
		cu_tk = cu_tk->next = CreateToken(SYM_ARGSTART, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy("(argument start)");
#endif

		//֐
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
		if(tk->buf == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			FreeToken(tk);
			return NULL;
		}
		str_lower(tk->buf);
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}

		// (
		if(pi->type != SYM_OPEN){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			FreeToken(tk);
			return NULL;
		}
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		
		//
		cu_tk = Expression(pi, cu_tk);

		// )
		if(pi->type != SYM_CLOSE){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			FreeToken(tk);
			return NULL;
		}
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}

		//֐ǉ
		cu_tk = cu_tk->next = tk;
		break;

	case SYM_VARIABLE:
		//ϐ
		if(pi->decl == TRUE || pi->ei->decl == TRUE){
			//ϐ錾
			cu_tk = cu_tk->next = CreateToken(SYM_DECLVARIABLE, pi->p);
		}else{
			cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		}
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
		if(cu_tk->buf == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		str_lower(cu_tk->buf);
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		break;

	case SYM_CONST:
		//l萔
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		s = alloc_copy_n(pi->p, pi->r - pi->p + 1);
		if(s == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}

		if(*s != '0' || *(s + 1) == '\0'){
			//10i
			cu_tk->i = atoi(s);

		}else if(*(s + 1) == 'x' || *(s + 1) == 'X'){
			//16i
			cu_tk->i = x2d(s + 2);
		}else{
			//8i
			cu_tk->i = o2d(s + 1);
		}
		mem_free(&s);

		if(GetToken(pi) == FALSE){
			return NULL;
		}
		break;

	case SYM_STRING:
		//萔
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		cu_tk->buf = alloc_copy_n(pi->p + 1, pi->r - pi->p - 1);
		if(cu_tk->buf == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
		if(GetToken(pi) == FALSE){
			return NULL;
		}

		if(pi->decl == TRUE || pi->type == SYM_EQ){
			//Ώ
			cu_tk->Type = SYM_ASSIGNSTRING;
			cu_tk->nfa = rege_compile(cu_tk->buf);
		}else{
#ifdef STRING_TO_NFA
			cu_tk->nfa = rege_compile(cu_tk->buf);
#endif
		}
		break;
	}
	return cu_tk;
}


/******************************************************************************

	Hash

	nbV

******************************************************************************/

static TOKEN *Hash(PARSEINFO *pi, TOKEN *cu_tk)
{
	cu_tk = Primary(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_HASHOPEN){
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		if(pi->type == SYM_HASHCLOSE){
			//nbVL[̎w肪ꍇ͋󕶎ǉ
			cu_tk = cu_tk->next = CreateToken(SYM_STRING, pi->p);
			if(cu_tk == NULL){
				Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
				return NULL;
			}
			cu_tk->buf = alloc_copy("");
			if(cu_tk->buf == NULL){
				Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
				return NULL;
			}
		}else{
			//nbVL[
			cu_tk = Assignment(pi, cu_tk);
			if(pi->type != SYM_HASHCLOSE){
				Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
				return NULL;
			}
		}
		//nbVʎq̒ǉ
		cu_tk = cu_tk->next = CreateToken(SYM_HASH, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy("(hash)");
#endif
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		cu_tk = Primary(pi, cu_tk);
		if(cu_tk == NULL){
			return NULL;
		}
	}
	return cu_tk;
}


/******************************************************************************

	UnaryOperator

	PZq

******************************************************************************/

static TOKEN *UnaryOperator(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Hash(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_NOT || pi->type == SYM_BITNOT ||
		pi->type == SYM_PLUS || pi->type == SYM_MINS ||
		pi->type == SYM_INC || pi->type == SYM_DEC ||
		pi->type == SYM_BINC || pi->type == SYM_BDEC){

		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(pi->type == SYM_BINC || pi->type == SYM_BDEC){
			//u
			pi->postfix = TRUE;
			cu_tk = cu_tk->next = tk;
			if(GetToken(pi) == FALSE){
				return NULL;
			}
			cu_tk = UnaryOperator(pi, cu_tk);
			if(cu_tk == NULL){
				return NULL;
			}
		}else{
			//Ou
			if(GetToken(pi) == FALSE){
				FreeToken(tk);
				return NULL;
			}
			cu_tk = UnaryOperator(pi, cu_tk);
			if(cu_tk == NULL){
				FreeToken(tk);
				return NULL;
			}
			cu_tk = cu_tk->next = tk;
		}
	}
	return cu_tk;
}


/******************************************************************************

	Multiplicative

	ZAZA]

******************************************************************************/

static TOKEN *Multiplicative(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = UnaryOperator(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_MULTI || pi->type == SYM_DIV || pi->type == SYM_MOD){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = UnaryOperator(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Additive

	ZAZ

******************************************************************************/

static TOKEN *Additive(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Multiplicative(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_ADD || pi->type == SYM_SUB){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Multiplicative(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Shift

	rbgVtg

******************************************************************************/

static TOKEN *Shift(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Additive(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_LEFTSHIFT || pi->type == SYM_RIGHTSHIFT){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Additive(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Relational

	r (<, >, <=, >=)

******************************************************************************/

static TOKEN *Relational(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Shift(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_LEFT || pi->type == SYM_LEFTEQ ||
		pi->type == SYM_RIGHT || pi->type == SYM_RIGHTEQ){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Shift(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Equality

	r (==, !=, =~)

******************************************************************************/

static TOKEN *Equality(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Relational(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_EQEQ || pi->type == SYM_REGEQ || pi->type == SYM_NTEQ){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Relational(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	AND

	_

******************************************************************************/

static TOKEN *AND(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = Equality(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_AND){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Equality(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	ExclusiveOR

	rI_a

******************************************************************************/

static TOKEN *ExclusiveOR(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = AND(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_XOR){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = AND(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	OR

	_a

******************************************************************************/

static TOKEN *OR(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = ExclusiveOR(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_OR){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = ExclusiveOR(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	LogicalAND

	_

******************************************************************************/

static TOKEN *LogicalAND(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = OR(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_CPAND){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = OR(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	LogicalOR

	_a

******************************************************************************/

static TOKEN *LogicalOR(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = LogicalAND(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_CPOR){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = LogicalAND(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Assignment

	

******************************************************************************/

static TOKEN *Assignment(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	cu_tk = LogicalOR(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_EQ){
		tk = CreateToken(pi->type, pi->p);
		if(tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = Assignment(pi, cu_tk);
		if(cu_tk == NULL){
			FreeToken(tk);
			return NULL;
		}
		cu_tk = cu_tk->next = tk;
	}
	return cu_tk;
}


/******************************************************************************

	Postfix

	u̎sʒu

******************************************************************************/

static TOKEN *Postfix(PARSEINFO *pi, TOKEN *cu_tk)
{
	cu_tk = Assignment(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	if(pi->postfix == FALSE){
		//㏈̕KvȂ
		pi->postfix = FALSE;
		return cu_tk;
	}
	pi->postfix = FALSE;

	cu_tk = cu_tk->next = CreateToken(SYM_POSTFIX, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("(postfix)");
#endif
	return cu_tk;
}


/******************************************************************************

	Expression

	1 , 2

******************************************************************************/

static TOKEN *Expression(PARSEINFO *pi, TOKEN *cu_tk)
{
	cu_tk = Postfix(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	while(pi->type == SYM_WORDEND){
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		cu_tk = Postfix(pi, cu_tk);
		if(cu_tk == NULL){
			return NULL;
		}
	}
	return cu_tk;
}


/******************************************************************************

	ExpressionStatement

	 ;

******************************************************************************/

static TOKEN *ExpressionStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	cu_tk = Expression(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	if(pi->type == SYM_EOF || pi->type == SYM_BCLOSE){
		return cu_tk;
	}
	if(pi->type != SYM_LINEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	Condition

	

******************************************************************************/

static TOKEN *Condition(PARSEINFO *pi, TOKEN *cu_tk)
{
	// (
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	if(pi->type != SYM_OPEN){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	if(pi->type == SYM_CLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	//
	cu_tk = Postfix(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	// )
	if(pi->type != SYM_CLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	pi->concat = TRUE;
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	IfStatement

	if

******************************************************************************/

static TOKEN *IfStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *end_tk, *else_tk;

	//
	cu_tk = Condition(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	//CMPǉ
	cu_tk = cu_tk->next = CreateToken(SYM_CMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("CMP");
#endif

	//ELSEɈړ
	else_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump ELSE");
#endif

	//Ώ
	cu_tk = StatementList(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}

	//rIʒuɈړ
	end_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump CMPEND");
#endif

	//ELSE:
	else_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_ELSE, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("ELSE:");
#endif
	if(pi->type == SYM_ELSE){
		//Ώ
		pi->concat = TRUE;
		if(GetToken(pi) == FALSE){
			return NULL;
		}
		if(pi->type == SYM_IF){
			cu_tk = IfStatement(pi, cu_tk);
		}else{
			cu_tk = StatementList(pi, cu_tk);
			if(cu_tk == NULL){
				return NULL;
			}
		}
	}

	//rIʒu
	end_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_CMPEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("CMPEND:");
#endif
	return cu_tk;
}


/******************************************************************************

	SwitchStatement

	switch

******************************************************************************/

static TOKEN *SwitchStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	//
	cu_tk = Condition(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	//switch̒ǉ
	cu_tk = cu_tk->next = CreateToken(SYM_SWITCH, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif

	//Ώ
	if(pi->type != SYM_BOPEN){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = CompoundStatement(pi, cu_tk);
	return cu_tk;
}


/******************************************************************************

	WhileStatement

	while

******************************************************************************/

static TOKEN *WhileStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *st_tk, *end_tk, ttk;

	//[vJnʒu
	st_tk = cu_tk = cu_tk->next = CreateToken(SYM_LOOPSTART, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPSTART:");
#endif

	//
	cu_tk = Condition(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	//LOOPǉ
	cu_tk = cu_tk->next = CreateToken(SYM_LOOP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOP");
#endif

	//Ώ
	ttk.next = NULL;
	if(StatementList(pi, &ttk) == NULL){
		FreeToken(ttk.next);
		return NULL;
	}
	cu_tk->target = ttk.next;

	//[vIʒuɈړ
	end_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPEND");
#endif

	//[vJnʒuɈړ
	cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPSTART");
#endif
	cu_tk->link = st_tk;

	//[vIʒu
	end_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_LOOPEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPEND:");
#endif
	return cu_tk;
}


/******************************************************************************

	DoWhileStatement

	do`while

******************************************************************************/

static TOKEN *DoWhileStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *st_tk, *end_tk, ttk;

	//[vJnʒu
	st_tk = cu_tk = cu_tk->next = CreateToken(SYM_LOOPSTART, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPSTART:");
#endif

	//LOOPǉ
	cu_tk = cu_tk->next = CreateToken(SYM_LOOP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOP");
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}

	//Ώ
	ttk.next = NULL;
	if(StatementList(pi, &ttk) == NULL){
		FreeToken(ttk.next);
		return NULL;
	}
	cu_tk->target = ttk.next;

	//LOOPp̃_~[
	cu_tk = cu_tk->next = CreateToken(SYM_DAMMY, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("DAMMY:");
#endif

	//whileĂ邩`FbN
	if(pi->type != SYM_WHILE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	//
	cu_tk = Condition(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	//CMPǉ
	cu_tk = cu_tk->next = CreateToken(SYM_CMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("CMP");
#endif

	if(pi->type != SYM_LINEEND && pi->type != SYM_EOF && pi->type != SYM_BCLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}

	//[vIʒuɈړ
	end_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPEND");
#endif

	//[vJnʒuɈړ
	cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPSTART");
#endif
	cu_tk->link = st_tk;

	//[vIʒu
	end_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_LOOPEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPEND:");
#endif
	return cu_tk;
}


/******************************************************************************

	ForStatement

	for

******************************************************************************/

static TOKEN *ForStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *st_tk, *end_tk, re_tk, ttk;

	if(GetToken(pi) == FALSE){
		return NULL;
	}

	// (
	if(pi->type != SYM_OPEN){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	if(GetToken(pi) == FALSE){
		return NULL;
	}

	//
	cu_tk = Expression(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	// ;
	if(pi->type != SYM_LINEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif

	//[vJnʒu
	st_tk = cu_tk = cu_tk->next = CreateToken(SYM_LOOPSTART, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPSTART:");
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}

	//
	cu_tk = Postfix(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}

	// ;
	if(pi->type != SYM_LINEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	if(GetToken(pi) == FALSE){
		return NULL;
	}

	//ď
	re_tk.next = NULL;
	if(Expression(pi, &re_tk) == NULL){
		return NULL;
	}

	// )
	if(pi->type != SYM_CLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		FreeToken(re_tk.next);
		return NULL;
	}

	//LOOPǉ
	cu_tk = cu_tk->next = CreateToken(SYM_LOOP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		FreeToken(re_tk.next);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOP");
#endif
	pi->concat = TRUE;
	if(GetToken(pi) == FALSE){
		FreeToken(re_tk.next);
		return NULL;
	}

	//Ώ
	ttk.next = NULL;
	if(StatementList(pi, &ttk) == NULL){
		FreeToken(re_tk.next);
		FreeToken(ttk.next);
		return NULL;
	}
	cu_tk->target = ttk.next;

	//[vIʒuɈړ
	end_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		FreeToken(re_tk.next);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPEND");
#endif

	//ďǉ
	for(cu_tk->next = re_tk.next; cu_tk->next != NULL; cu_tk = cu_tk->next);
	cu_tk = cu_tk->next = CreateToken(SYM_LINEEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy(";");
#endif

	//[vJnʒuɈړ
	cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump LOOPSTART");
#endif
	cu_tk->link = st_tk;

	//[vIʒu
	end_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_LOOPEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("LOOPEND:");
#endif
	return cu_tk;
}


/******************************************************************************

	JumpStatement

	Wv

******************************************************************************/

static TOKEN *JumpStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk;

	//token
	tk = CreateToken(pi->type, pi->p);
	if(tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		FreeToken(tk);
		return NULL;
	}
	
	//l
	cu_tk = Postfix(pi, cu_tk);
	if(cu_tk == NULL){
		FreeToken(tk);
		return NULL;
	}
	cu_tk = cu_tk->next = tk;

	if(pi->type == SYM_EOF || pi->type == SYM_BCLOSE){
		return cu_tk;
	}
	if(pi->type != SYM_LINEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	LabeldStatement

	x

******************************************************************************/

static TOKEN *LabeldStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	int type = pi->type;

	cu_tk = cu_tk->next = CreateToken(type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	if(type == SYM_CASE){
		//l
		cu_tk = Postfix(pi, cu_tk);
		if(cu_tk == NULL){
			return NULL;
		}
	}
	if(pi->type != SYM_CASEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	VarDeclList

	ϐXg

******************************************************************************/

static TOKEN *VarDeclList(PARSEINFO *pi, TOKEN *cu_tk)
{
	while(1){
		//
		pi->decl = TRUE;
		cu_tk = Hash(pi, cu_tk);
		pi->decl = FALSE;
		if(cu_tk == NULL){
			return NULL;
		}

		if(pi->type == SYM_EQ){
			// = 
			cu_tk = Postfix(pi, cu_tk);
			if(cu_tk == NULL){
				return NULL;
			}
		}

		if(pi->type != SYM_WORDEND){
			break;
		}
		// ,
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			return NULL;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		if(GetToken(pi) == FALSE){
			return NULL;
		}
	}
	return cu_tk;
}


/******************************************************************************

	VarDecl

	ϐ

******************************************************************************/

static TOKEN *VarDecl(PARSEINFO *pi, TOKEN *cu_tk)
{
	//var
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	//ϐXg
	cu_tk = VarDeclList(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}
	if(pi->type == SYM_EOF || pi->type == SYM_BCLOSE){
		return cu_tk;
	}
	if(pi->type != SYM_LINEEND){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	// ;
	cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	FuncDecl

	֐

******************************************************************************/

static TOKEN *FuncDecl(PARSEINFO *pi, TOKEN *cu_tk)
{
	FUNCINFO *fi;
	TOKEN *tk, *end_tk;
	int type = pi->type;

	//֐IʒuɈړ
	end_tk = cu_tk = cu_tk->next = CreateToken(SYM_JUMP, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("jump FUNCEND");
#endif

	//֐ ( Function | ArgMatch )
	cu_tk = cu_tk->next = CreateToken(type, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
	if(GetToken(pi) == FALSE){
		return NULL;
	}

	//֐
	if(pi->type != SYM_FUNC){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	tk = CreateToken(SYM_FUNC, pi->p);
	if(tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
	tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
	if(tk->buf == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	str_lower(tk->buf);
	if(GetToken(pi) == FALSE){
		FreeToken(tk);
		return NULL;
	}

	//Ăяop̐ݒ
	if(pi->ei->sci->fi == NULL){
		fi = pi->ei->sci->fi = mem_alloc(sizeof(FUNCINFO));
	}else{
		for(fi = pi->ei->sci->fi; fi->next != NULL; fi = fi->next);
		fi->next = mem_alloc(sizeof(FUNCINFO));
		fi = fi->next;
	}
	if(fi == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	ZeroMemory(fi, sizeof(FUNCINFO));
	fi->Name = alloc_copy(tk->buf);
	if(fi->Name == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	fi->name_hash = str2hash(fi->Name);
	fi->tk = cu_tk;

	// (
	if(pi->type != SYM_OPEN){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	if(GetToken(pi) == FALSE){
		FreeToken(tk);
		return NULL;
	}

	//
	if(type == SYM_ARGMATCH && pi->type != SYM_STRING){
		//ArgMatch֐̃G[`FbN
		Error(pi->ei, ERR_ARGFORMAT, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	cu_tk = VarDeclList(pi, cu_tk);

	// )
	if(pi->type != SYM_CLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		FreeToken(tk);
		return NULL;
	}
	pi->concat = TRUE;
	if(GetToken(pi) == FALSE){
		FreeToken(tk);
		return NULL;
	}

	//֐̐ݒ
	cu_tk = cu_tk->next = tk;

	//֐{
	if(pi->type != SYM_BOPEN){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	cu_tk = CompoundStatement(pi, cu_tk);
	if(cu_tk == NULL){
		return NULL;
	}

	//֐Iʒu
	end_tk->link = cu_tk = cu_tk->next = CreateToken(SYM_FUNCEND, pi->p);
	if(cu_tk == NULL){
		Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
		return NULL;
	}
#ifdef DEBUG_SET
	cu_tk->buf = alloc_copy("FUNCEND:");
#endif
	return cu_tk;
}


/******************************************************************************

	CompoundStatement

	ubN

******************************************************************************/

static TOKEN *CompoundStatement(PARSEINFO *pi, TOKEN *cu_tk)
{
	TOKEN *tk, ttk;

	if(GetToken(pi) == FALSE){
		return NULL;
	}
	//ubN̏Iʒu܂ŉ
	ttk.next = NULL;
	tk = &ttk;
	while(tk != NULL && pi->type != SYM_EOF && pi->type != SYM_BCLOSE){
		tk = StatementList(pi, tk);
	}
	cu_tk->target = ttk.next;
	if(tk == NULL){
		return NULL;
	}
	if(pi->type != SYM_BCLOSE){
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		return NULL;
	}
	pi->concat = TRUE;
	if(GetToken(pi) == FALSE){
		return NULL;
	}
	return cu_tk;
}


/******************************************************************************

	StatementList

	

******************************************************************************/

static TOKEN *StatementList(PARSEINFO *pi, TOKEN *cu_tk)
{
	pi->level++;

	switch(pi->type)
	{
	case SYM_EOF:
		break;

	case SYM_PREP:
		//vvZbT
		pi->r = Preprocessor(pi->ei->sci, pi->ei->sci->path, pi->r);
		if(pi->r == NULL){
			cu_tk = NULL;
			break;
		}
		if(GetToken(pi) == FALSE){
			cu_tk = NULL;
			break;
		}
		break;

	case SYM_BOPEN:
		//ubN
		cu_tk = cu_tk->next = CreateToken(pi->type, pi->p);
		if(cu_tk == NULL){
			Error(pi->ei, ERR_ALLOC, pi->p, FALSE);
			break;
		}
#ifdef DEBUG_SET
		cu_tk->buf = alloc_copy_n(pi->p, pi->r - pi->p + 1);
#endif
		cu_tk = CompoundStatement(pi, cu_tk);
		break;

	case SYM_IF:
		cu_tk = IfStatement(pi, cu_tk);
		break;

	case SYM_SWITCH:
		cu_tk = SwitchStatement(pi, cu_tk);
		break;

	case SYM_WHILE:
		cu_tk = WhileStatement(pi, cu_tk);
		break;

	case SYM_DO:
		cu_tk = DoWhileStatement(pi, cu_tk);
		break;

	case SYM_FOR:
		cu_tk = ForStatement(pi, cu_tk);
		break;

	case SYM_RETURN:
	case SYM_EXIT:
	case SYM_BREAK:
	case SYM_CONTINUE:
		cu_tk = JumpStatement(pi, cu_tk);
		break;

	case SYM_CASE:
	case SYM_DEFAULT:
		cu_tk = LabeldStatement(pi, cu_tk);
		break;

	case SYM_VAR:
		//ϐ
		cu_tk = VarDecl(pi, cu_tk);
		break;

	case SYM_FUNCTION:
	case SYM_ARGMATCH:
		//֐
		if(pi->level > 1){
			Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
			cu_tk = NULL;
			break;
		}
		cu_tk = FuncDecl(pi, cu_tk);
		break;

	case SYM_FUNC:
	case SYM_VARIABLE:
	case SYM_CONST:
	case SYM_STRING:
	case SYM_OPEN:
	case SYM_NOT:
	case SYM_BITNOT:
	case SYM_PLUS:
	case SYM_MINS:
	case SYM_INC:
	case SYM_DEC:
	case SYM_WORDEND:
	case SYM_LINEEND:
		//
		cu_tk = ExpressionStatement(pi, cu_tk);
		break;

	default:
		Error(pi->ei, ERR_SENTENCE, pi->p, FALSE);
		cu_tk = NULL;
		break;
	}
	pi->level--;
	return cu_tk;
}


/******************************************************************************

	ParseVariable

	ϐ擾

******************************************************************************/

TOKEN *ParseVariable(EXECINFO *ei, char *buf)
{
	TOKEN tk;
	PARSEINFO pi;

	ZeroMemory(&pi, sizeof(PARSEINFO));
	pi.ei = ei;
	pi.p = pi.r = buf;
	pi.concat = TRUE;
	pi.level = 1;

	if(GetToken(&pi) == FALSE){
		return NULL;
	}

	//
	ZeroMemory(&tk, sizeof(TOKEN));
	if(Hash(&pi, &tk) == NULL || pi.type != SYM_EOF){
		//G[͑Sĉ
		FreeToken(tk.next);
		FreeFuncInfo(ei->sci->fi);
		ei->sci->fi = NULL;
		return NULL;
	}
	return tk.next;
}


/******************************************************************************

	ParseSentence

	\

******************************************************************************/

TOKEN *ParseSentence(EXECINFO *ei, char *buf, int level)
{
	TOKEN *cu_tk, tk;
	PARSEINFO pi;

	ZeroMemory(&pi, sizeof(PARSEINFO));
	pi.ei = ei;
	pi.p = pi.r = buf;
	pi.concat = TRUE;
	pi.level = level;

	if(GetToken(&pi) == FALSE){
		return NULL;
	}

	//
	ZeroMemory(&tk, sizeof(TOKEN));
	cu_tk = &tk;
	while(cu_tk != NULL && pi.type != SYM_EOF){
		cu_tk = StatementList(&pi, cu_tk);
	}
	if(cu_tk == NULL){
		//G[͑Sĉ
		FreeToken(tk.next);
		FreeFuncInfo(ei->sci->fi);
		ei->sci->fi = NULL;
		return NULL;
	}
	return tk.next;
}
/* End of source */
